<?php
/*
 * Plugin Name: BastionBytes integration for BookStack
 * Description: simple BookStack integration for wordpress
 * Version: 0.4.4
 * Author: Martin Dames (martin@bastionbytes.de)
 * License: GPLv3
 * Text Domain: bastionbytes-integration-for-bookstack
 * Domain Path: /languages
 */

if(!defined('ABSPATH')) {
	// direct access
	die();
}

// used to pass messages to the wp backend page
$bb_integration_for_bookstack_plugin_message_error = false;
$bb_integration_for_bookstack_plugin_message_success = false;

// query remote bookstack (returns false on any failure)
function bb_integration_for_bookstack_query_bookstack($id) {

	$bookstack_token_id = get_option('bb_integration_for_bookstack_token_id');
	$bookstack_token_secret = get_option('bb_integration_for_bookstack_token_secret');
	$bookstack_url = get_option('bb_integration_for_bookstack_url');

	$query_options = [
		'timeout' => 3,
		'headers' => 'Authorization: Token ' . $bookstack_token_id . ':' . $bookstack_token_secret
	];

	if(get_option('bb_integration_for_bookstack_ignore_cert') == 1) $query_options['sslverify'] = false;

	$result = wp_remote_get($bookstack_url . "/api/pages/" . $id, $query_options);

	if(is_wp_error($result)) {
		return(false);
	}

	$result = json_decode(wp_remote_retrieve_body($result));
	if(is_null($result)) return(false);

	if(!property_exists($result, 'name') || !property_exists($result, 'html')) return(false);

	$r = array(
		'title' => $result->name,
		'content' => $result->html
	);
	return($r);
}

// try to get bookstack entry from cache
function bb_integration_for_bookstack_query_db($id) {
	global $wpdb;
	$result = $wpdb->get_row($wpdb->prepare(
		"SELECT `valid`, UNIX_TIMESTAMP(NOW()) - UNIX_TIMESTAMP(`updated`) AS `age`, `title`, `content` " .
		"FROM `" . $wpdb->prefix . "bb_integration_for_bookstack` WHERE `id`=%d", $id));
	if(is_null($result)) return(false);
	return($result);
}

// format output
function bb_integration_for_bookstack_generate_entry($title, $content, $showTitle, $hideTitle) {

	if(!$showTitle && !$hideTitle) {
		$showTitle = get_option('bb_integration_for_bookstack_show_title') == 1;
	} else if($hideTitle) {
		$showTitle = false;
	}

	$r = "";
	if($showTitle) {
		$r .= "<div class='bb-bookstack-title'>" . $title . "</div>";
	}
	$r .= "<div class='bb-bookstack-content'>" . $content . "</div>";

	return($r);
}

// update or create cache entry
function bb_integration_for_bookstack_update_db($id, $title, $content, $createEntry) {
	global $wpdb;
	$query = false;
	if($createEntry) {
		$wpdb->query($wpdb->prepare("INSERT IGNORE INTO `" . $wpdb->prefix . "bb_integration_for_bookstack` (`id`, `title`, `content`) VALUES (%d, %s, %s)", $id, $title, $content));
	} else {
		$wpdb->query($wpdb->prepare("UPDATE `" . $wpdb->prefix . "bb_integration_for_bookstack` SET `title`=%s, `content`=%s, `valid`=1, `updated`=NOW() WHERE `id`=%d", $title, $content, $id));
	}
}

// the registered shortcode function
function bb_integration_for_bookstack_shortcode($atts = [], $content = null) {

	if(!is_numeric($atts['id'])) return(__('Content cannot be loaded.', 'bastionbytes-integration-for-bookstack'));
	$id = $atts['id'] + 0; // ensure that this is an integer

	$showTitle = in_array('show_title', $atts);
	$hideTitle = in_array('hide_title', $atts);

	if($showTitle && $hideTitle) {
		$hideTitle = false;
	}

	$mode = get_option('bb_integration_for_bookstack_mode');
	$maxCacheAge = get_option('bb_integration_for_bookstack_max_cache_age');

	$dbResult = false;
	if($mode != 'no_cache') $dbResult = bb_integration_for_bookstack_query_db($id);

	if(($dbResult !== false) && $dbResult->valid && ($dbResult->age < $maxCacheAge * 3600)) {
		// cache hit
		return(bb_integration_for_bookstack_generate_entry($dbResult->title, $dbResult->content, $showTitle, $hideTitle));
	}

	// cache miss or cache entry invalid
	$bsResult = false;
	if($mode != 'no_bookstack') $bsResult = bb_integration_for_bookstack_query_bookstack($id);

	if($bsResult === false) {
		if($dbResult !== false) {
			// unable to connect to bookstack => just return the old cached entry
			return(bb_integration_for_bookstack_generate_entry($dbResult->title, $dbResult->content, $showTitle, $hideTitle));
		}
		return(__('Content cannot be loaded.', 'bastionbytes-integration-for-bookstack'));
	}

	// update cache entry
	bb_integration_for_bookstack_update_db($id, $bsResult['title'], $bsResult['content'], $dbResult === false);

	// return formated entry
	return(bb_integration_for_bookstack_generate_entry($bsResult['title'], $bsResult['content'], $showTitle, $hideTitle));

}

function bb_integration_for_bookstack_generate_webhook_secret() {
	$cs = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
	$csLen = strlen($cs);
	$r = '';
	for($i = 0; $i < 32; $i++) {
		$r .= $cs[random_int(0, $csLen - 1)];
	}
	return($r);
}

// registered plugin init function
function bb_integration_for_bookstack_init() {
	global $wpdb;

	add_option('bb_integration_for_bookstack_url', 'https://url.to.your.bookstack');
	add_option('bb_integration_for_bookstack_token_id', 'your_token_id');
	add_option('bb_integration_for_bookstack_token_secret', 'your_token_secret');
	add_option('bb_integration_for_bookstack_webhook_secret', bb_integration_for_bookstack_generate_webhook_secret());
	add_option('bb_integration_for_bookstack_mode', 'normal');
	add_option('bb_integration_for_bookstack_max_cache_age', 24);
	add_option('bb_integration_for_bookstack_irgnore_cert', 0);

	// id: the bookstack page id
	// valid: if the cache entry is valid
	// updated: last renew timestamp of the entry
	// title: the page title
	// content: the page content

	$query = "CREATE TABLE `" . $wpdb->prefix . "bb_integration_for_bookstack` (
		`id` INT NOT NULL UNIQUE,
		`valid` INT(1) NOT NULL DEFAULT '1',
		`updated` TIMESTAMP NOT NULL DEFAULT NOW(),
		`title` TEXT NOT NULL,
		`content` TEXT NOT NULL
		) " . $wpdb->get_charset_collate() . ";";

	require_once( ABSPATH . 'wp-admin/includes/upgrade.php' );
	dbDelta($query);
}

// the backend configuration page
function bb_integration_for_bookstack_backend_options() {
	global $bb_integration_for_bookstack_plugin_message_error, $bb_integration_for_bookstack_plugin_message_success;
?>
	<div class='wrap'>
		<h1><?php esc_html_e('Settings > BastionBytes integration for BookStack', 'bastionbytes-integration-for-bookstack'); ?></h1>
		<?php if($bb_integration_for_bookstack_plugin_message_error !== false) { ?>
			<div id="bookstack-error" class="notice notice-error settings-error">
				<p>
					<strong><?php echo esc_html($bb_integration_for_bookstack_plugin_message_error); ?></strong>
				</p>
			</div>
		<?php } ?>
		<?php if($bb_integration_for_bookstack_plugin_message_success !== false) { ?>
			<div id="bookstack-error-settings_updated" class="notice notice-success settings-error"> 
				<p>
					<strong><?php echo esc_html($bb_integration_for_bookstack_plugin_message_success); ?></strong>
				</p>
			</div>
		<?php } ?>
		<form action="<?php menu_page_url('bb_integration_for_bookstack_backend_options'); ?>" method="post">
			<?php wp_nonce_field('bb_integration_for_bookstack_backend_options'); ?>
			<table class="form-table" role="presentation">
				<tr>
					<th scope="row">
						<label><?php esc_html_e('BookStack URL', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<input type="text" name="bb_integration_for_bookstack_option_url" class="regular-text" value="<?php echo esc_html(get_option('bb_integration_for_bookstack_url')); ?>">
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('BookStack Token ID', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<input type="text" name="bb_integration_for_bookstack_option_token_id" class="regular-text" value="<?php echo esc_html(get_option('bb_integration_for_bookstack_token_id')); ?>">
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('BookStack Token Secret', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<input type="text" name="bb_integration_for_bookstack_option_token_secret" class="regular-text" value="" placeholder="<?php esc_html_e('hidden', 'bastionbytes-integration-for-bookstack'); ?>">
						<p class="description">
							<?php esc_html_e('If this field is left empty, no changes will be made', 'bastionbytes-integration-for-bookstack'); ?>
						</p>
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('Show BookStack page title', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<?php
							$checked = "";
							if(get_option('bb_integration_for_bookstack_show_title') == 1) $checked = "checked";
						?>
						<input name="bb_integration_for_bookstack_option_show_title" type="checkbox" <?php echo esc_html($checked); ?>>
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('Cache-Timeout (hours)', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<input type="number" name="bb_integration_for_bookstack_option_max_cache_age" class="regular-text" value="<?php echo esc_html(get_option('bb_integration_for_bookstack_max_cache_age')); ?>">
						<p class="description">
							<?php esc_html_e('The value &quot;0&quot; means that cache entries never expire', 'bastionbytes-integration-for-bookstack'); ?>
						</p>
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('Mode', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<?php
							$mode = get_option('bb_integration_for_bookstack_mode');
							$modes = ['normal', 'no_bookstack', 'no_cache'];
							$mode_labels = [
								__('normal', 'bastionbytes-integration-for-bookstack'),
								__('never connect to BookStack (use cache only)', 'bastionbytes-integration-for-bookstack'),
								__('always connect to BookStack (turn off cache)', 'bastionbytes-integration-for-bookstack')
							];
							if(!in_array($mode, $modes)) $mode = 'normal'
						?>
						<select name="bb_integration_for_bookstack_option_mode">
							<?php
								for($i = 0; $i < count($modes); $i++) {
									echo '<option value="' . esc_html($modes[$i]) . '"';
									if($modes[$i] == $mode) echo ' selected';
									echo '>' . esc_html($mode_labels[$i]) . '</option>';
								}
							?>
						</select>
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('Ignore BookStack certificate', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<?php
							$checked = "";
							if(get_option('bb_integration_for_bookstack_ignore_cert') == 1) $checked = "checked";
						?>
						<input name="bb_integration_for_bookstack_option_ignore_cert" type="checkbox" <?php echo esc_html($checked); ?>>
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('BookStack Webhook URL', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td><?php echo esc_html(get_site_url()); ?>/wp-json/bookstack/v1/site-update?<?php echo esc_html(get_option('bb_integration_for_bookstack_webhook_secret')); ?></td>
				</tr>
			</table>
			<p class="submit">
				<input type="submit" name="bb_integration_for_bookstack_update_config" class="button button-primary" value="<?php esc_html_e('Save changes', 'bastionbytes-integration-for-bookstack'); ?>">
			</p>

			<hr style='border-bottom-width: 2px;border-top-width: 2px;border-top-style: solid;'>

			<h2><?php esc_html_e('Administrativ action', 'bastionbytes-integration-for-bookstack'); ?></h2>
			<table class="form-table" role="presentation">
				<tr>
					<th scope="row">
						<label><?php esc_html_e('Regenerate Webhook URL', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<input type="submit" name="bb_integration_for_bockstack_generate_webhook_secret" class="button button-primary" value="<?php esc_html_e('Execute action', 'bastionbytes-integration-for-bookstack'); ?>">
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('Invalidate all cache entries', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<input type="submit" name="bb_integration_for_bookstack_invalidate_cache" class="button button-primary" value="<?php esc_html_e('Execute action', 'bastionbytes-integration-for-bookstack'); ?>">
					</td>
				</tr>
				<tr>
					<th scope="row">
						<label><?php esc_html_e('Delete all cache entries', 'bastionbytes-integration-for-bookstack'); ?></label>
					</th>
					<td>
						<input type="submit" name="bb_integration_for_bookstack_drop_cache" class="button button-primary" value="<?php esc_html_e('Execute action', 'bastionbytes-integration-for-bookstack'); ?>">
					</td>
				</tr>
			</table>
		</form>
		<hr style='border-bottom-width: 2px;border-top-width: 2px;border-top-style: solid;'>
		<?php
			$helpFile = realpath(dirname(__FILE__)) . '/languages/help-' . get_locale() . '.html';
			if(!file_exists($helpFile)) {
				$helpFile = realpath(dirname(__FILE__)) . '/languages/help.html';
			}
			include($helpFile);
		?>
	</div>
<?php
}

// the missing php startsWith function for strings
function bb_integration_for_bookstack_starts_with($x, $prefix) {
	if(strlen($x) < strlen($prefix)) {
		return(true);
	}
	return(substr($x, 0, strlen($prefix)) == $prefix);
}

// the form submit function registered in bookstack_backend_options_menu()
function bb_integration_for_bookstack_backend_options_submit() {
	global $bb_integration_for_bookstack_plugin_message_error, $bb_integration_for_bookstack_plugin_message_success, $wpdb;

	if(!isset($_SERVER['REQUEST_METHOD'])) return;
	if($_SERVER['REQUEST_METHOD'] !== 'POST') return;

	// check nonce
	check_admin_referer('bb_integration_for_bookstack_backend_options');

	if(isset($_POST['bb_integration_for_bookstack_generate_webhook_secret'])) {

		update_option('bb_integration_for_bookstack_webhook_secret', bb_integration_for_bookstack_generate_webhook_secret());
		$bb_integration_for_bookstack_plugin_message_success = __('New webhook URL created', 'bastionbytes-integration-for-bookstack');

	} else if(isset($_POST['bb_integration_for_bookstack_invalidate_cache'])) {

		$wpdb->query("UPDATE `" . $wpdb->prefix . "bb_integration_for_bookstack` SET `valid`=0");
		$bb_integration_for_bookstack_plugin_message_success = __('All cache entries invalidated', 'bastionbytes-integration-for-bookstack');

	} else if(isset($_POST['bb_integration_for_bookstack_drop_cache'])) {

		$wpdb->query("DELETE FROM `" . $wpdb->prefix . "bb_integration_for_bookstack`");
		$bb_integration_for_bookstack_plugin_message_success = __('All cache entries deleted', 'bastionbytes-integration-for-bookstack');

	} else if(isset($_POST['bb_integration_for_bookstack_update_config'])) {

		$url = '';
		$token_id = '';
		$token_secret = '';
		$max_cache_age = '';
		if(isset($_POST['bb_integration_for_bookstack_option_url'])) $url = sanitize_url(wp_unslash($_POST['bb_integration_for_bookstack_option_url']));
		if(isset($_POST['bb_integration_for_bookstack_option_token_id'])) $token_id = trim(sanitize_text_field(wp_unslash($_POST['bb_integration_for_bookstack_option_token_id'])));
		if(isset($_POST['bb_integration_for_bookstack_option_token_secret'])) $token_secret = trim(sanitize_text_field(wp_unslash($_POST['bb_integration_for_bookstack_option_token_secret'])));
		if(isset($_POST['bb_integration_for_bookstack_option_max_cache_age'])) $max_cache_age = trim(sanitize_text_field(wp_unslash($_POST['bb_integration_for_bookstack_option_max_cache_age'])));

		$mode = 'normal';
		$modes = ['normal', 'no_bookstack', 'no_cache'];
		if(isset($_POST['bb_integration_for_bookstack_option_mode'])) $mode = trim(sanitize_text_field(wp_unslash($_POST['bb_integration_for_bookstack_option_mode'])));
		if(!in_array($mode, $modes)) $mode = 'normal';

		// some input validation checks
		if(!bb_integration_for_bookstack_starts_with($url, "https://") && !bb_integration_for_bookstack_starts_with($url, "http://")) {
			$bb_integration_for_bookstack_plugin_message_error = __('The BookStack URL has to begin with http:// oder https://', 'bastionbytes-integration-for-bookstack');
		}
		$host = substr($url, strpos($url, '//') + 2);
		if(strlen($host) < 4 || strlen($host) > 100) {
			$bb_integration_for_bookstack_plugin_message_error = __('The BookStack URL is too short or too long', 'bastionbytes-integration-for-bookstack');
		}
		if(!preg_match('/^[a-zA-Z0-9\/.-]*$/', $host)) {
			$bb_integration_for_bookstack_plugin_message_error = __('The BookStack URL contains invalid characters', 'bastionbytes-integration-for-bookstack');
		}
		if(!preg_match('/^[a-zA-Z0-9]*$/', $token_id)) {
			$bb_integration_for_bookstack_plugin_message_error = __('The token ID contains invalid characters', 'bastionbytes-integration-for-bookstack');
		}
		if(strlen($token_id) < 4 || strlen($token_id) > 100) {
			$bb_integration_for_bookstack_plugin_message_error = __('The token ID is too short or too long', 'bastionbytes-integration-for-bookstack');
		}
		if($token_secret != "") {
			if(!preg_match('/^[a-zA-Z0-9]*$/', $token_secret)) {
				$bb_integration_for_bookstack_plugin_message_error = __('The token secret contains invalid characters', 'bastionbytes-integration-for-bookstack');
			}
			if(strlen($token_secret) < 4 || strlen($token_secret) > 100) {
				$bb_integration_for_bookstack_plugin_message_error = __('The token secret is too short or too long', 'bastionbytes-integration-for-bookstack');
			}
		}
		if(!is_numeric($max_cache_age)) {
			$bb_integration_for_bookstack_plugin_message_error = __('Cache-Timeout is not a number', 'bastionbytes-integration-for-bookstack');
		}
		$max_cache_age = intval($max_cache_age);
		if(($max_cache_age < 0) || ($max_cache_age > 336)) {
			$bb_integration_for_bookstack_plugin_message_error = __('Cache-Timeout must be at least 0 and maximum of 336', 'bastionbytes-integration-for-bookstack');
		}



		// if no error occured update the values
		if($bb_integration_for_bookstack_plugin_message_error === false) {

			while(substr($url, -1) == '/') {
				$url = substr($url, 0, -1);
			}

			update_option('bb_integration_for_bookstack_url', $url);
			update_option('bb_integration_for_bookstack_token_id', $token_id);
			update_option('bb_integration_for_bookstack_max_cache_age', $max_cache_age);

			$show_title = 0;
			if(isset($_POST['bb_integration_for_bookstack_option_show_title'])) $show_title = 1;
			update_option('bb_integration_for_bookstack_show_title', $show_title);

			$ignore_cert = 0;
			if(isset($_POST['bb_integration_for_bookstack_option_ignore_cert'])) $ignore_cert = 1;
			update_option('bb_integration_for_bookstack_ignore_cert', $ignore_cert);

			update_option('bb_integration_for_bookstack_mode', $mode);

			if($token_secret != "") {
				update_option('bb_integration_for_bookstack_token_secret', $token_secret);
			}
			$bb_integration_for_bookstack_plugin_message_success = __('Changes saved', 'bastionbytes-integration-for-bookstack');
		}
	}
}

// registered menu function
function bb_integration_for_bookstack_backend_options_menu() {
	$hook = add_submenu_page(
		'options-general.php',
		'BastionBytes integration for BookStack',
		'BastionBytes integration for BookStack',
		'manage_options',
		'bb_integration_for_bookstack_backend_options',
		'bb_integration_for_bookstack_backend_options'
	);
	add_action('load-' . $hook, 'bb_integration_for_bookstack_backend_options_submit');
}

// called after the plugin is loaded
function bb_integration_for_bookstack_plugin_loaded() {
	load_plugin_textdomain('bb_integration_for_bookstack', false, plugin_basename(dirname(__FILE__)) . '/languages');
}

function bb_integration_for_bookstack_webhook(WP_REST_Request $request) {
	global $wpdb;
	if(!isset($request[get_option('bb_integration_for_bookstack_webhook_secret')])) {
		http_response_code(401);
		die('permission denied');
	}
	$json = $request->get_json_params();
	if(!isset($json['related_item']['id']) || !isset($json['event']) || ($json['event'] != 'page_update')) {
		http_response_code(400);
		die('bad request');
	}
	$id = $json['related_item']['id'];
	if(!is_numeric($id)) {
		http_response_code(400);
		die('bad request');
	}

	$wpdb->query($wpdb->prepare('UPDATE `' . $wpdb->prefix . 'bb_integration_for_bookstack` SET `valid`=0 WHERE `id`=%d', $id));
}

add_shortcode('bb-bookstack', 'bb_integration_for_bookstack_shortcode');
register_activation_hook(__FILE__, 'bb_integration_for_bookstack_init');
add_action('admin_menu', 'bb_integration_for_bookstack_backend_options_menu');
add_action('plugins_loaded', 'bb_integration_for_bookstack_plugin_loaded');
add_action('rest_api_init', function () {
	register_rest_route('bookstack/v1', '/site-update', [
		'methods' => 'POST',
		'callback' => 'bb_integration_for_bookstack_webhook',
		'permission_callback' => '__return_true'
	]);
});

